{N.Yu.Zolotykh 1999}
{University of Nizhni Novgorod, Russia}

unit BigArith;
interface

{----------------------------------------------------------------------------}

type
  PWordArray = ^TWordArray;
    {you should not use directly this type in your own programs}
  TWordArray = Array[0..$FFFF div SizeOf(Word) - 1] of Word;
    {you should not use directly this type in your own programs}

  PBigStruct = ^TBigStruct;
    {you should not use directly this type in your own programs}
  TBigStruct = object
  private
    Sign: ShortInt;     {the sign: 0, 1 or -1}
    Data: PWordArray;   {the storage for digits}
    Len: Word;          {the number of digits}
    MemLen: Word;       {the size of storage: Len or Len+1}
    Refs: Word;         {the number of points to this number}
  end;
    {you should not use directly this type in your own programs}

{----------------------------------------------------------------------------}

  {BigInt: a big integer number type}

type

  BigInt = object {a Big Integer Number implementation}
    constructor Init;
      {initialize a BigInt}
      {you should always call this constructor befor using a BigInt}
    destructor Done;
      {destroy a BigInt}
    procedure AssignZero;
      {assign 0 to a BigInt}
    procedure AssignOne;
      {assign 1 to a BigInt}
    procedure AssignMinusOne;
      {assign -1 to a BigInt}
    procedure Assign(B: BigInt);
      {assign a value B to a BigInt}
    procedure AssignString(S: String);
      {convert a string to a BigInt}
      {this procedure calls Warning if S is emty or S contains non-digits}
    procedure AssignWord(W: Word);
      {convert a word to a BigInt}
    function IsZero: boolean;
      {return true if a BigInt is zero}
    function IsPos: boolean;
      {return true if a BigInt is positive}
    function IsNeg: boolean;
      {return true if a BigInt is negative}
    function ToString(var S: String): Boolean;
      {convert a BigInt to a String}
      {this function returns false and call Warning if a BigInt is too
      big to be allocated in a String}
    function ToWord(var W: Word): Boolean;
      {convert a BigInt to a word}
      {this function returns false and call Warning if a BigInt is too
      big to be allocated in a word}
    procedure BigIntRead(var F: Text);
      {read a number using ReadString and AssignString}
      {true if operation is done correctly}
    procedure BigIntWrite(var F: Text);
      {write a number using WriteString}
  private
    Number: PBigStruct;
    procedure _Free;
    procedure _FillGetMem(ASign: ShortInt; ALen, AMemLen: Word);
    procedure _NewNumber(ASign: ShortInt; AData:PWordArray; ALen, AMemLen: Word);
  end;

  procedure Add(var A: BigInt; B, C: BigInt);
    {A := B + C}
  procedure Sub(var A: BigInt; B, C: BigInt);
    {A := B - C}
  procedure Mult(var A: BigInt; B, C: BigInt);
    {A := B * C}
  procedure Divide(var A: BigInt; B, C: BigInt; var Res: BigInt);
    {A := B div C; Res := B mod C; B = A * C + Res}
    {if C = 0 calls FatalError('Divide by zero')}
  procedure Dvd(var A: BigInt; B, C: BigInt);
    {A := B div C}
    {if C = 0 calls FatalError('Divide by zero')}
  procedure Res(var R: BigInt; B, C: BigInt);
    {R := B mod C}
    {if C = 0 calls FatalError('Divide by zero')}
  procedure MultByWord(var A: BigInt; B: BigInt; C: Word);
    {A := B * C}
  procedure DivideByWord(var A: BigInt; B: BigInt; C: Word; var Res: Word);
    {A := B div C; Res := B mod C; 0 <= Res < C}
    {For positive B it holds B = A * C + Res}
    {For negative B it holds B = A * C - Res}
    {if C = 0 calls FatalError('Divide by zero')}
  procedure Minus(var A: BigInt; B: BigInt);
    {A := -B}
  function Cmp(A, B: BigInt): ShortInt;
    {return Sign(A-B): 0 if A = B, 1 if A > B, -1 if A < B}
  function Less(A, B: BigInt): Boolean;
    {return true if A < B}
  function Equal(A, B: BigInt): Boolean;
    {return true if A = B}
  function Greater(A, B: BigInt): Boolean;
    {return true if A > B}
  procedure AbsValue(var A: BigInt; B: BigInt);

var Zero, One, Two, MinusOne: BigInt;

type
  WriteProcedureType = procedure (var F: Text; S: String);
  ReadProcedureType = procedure (var F: Text; var S: String);

var
  StdErr: ^Text;
    {pointer to control stream of error/warning messages
    by default StdErr = @Output}

var
  Error: WriteProcedureType;
    {BigArith always calls this procedure with parameter F = StdErr^
    if an error has occured,
    by default it writes the message S in file F}
  FatalError: WriteProcedureType;
    {BigArith always calls this procedure with parameter F = StdErr^
    if a fatal error has occured,
    by default it writes the message S in the file F
    and terminates running the program}
  Warning: WriteProcedureType;
    {BigArith calls this procedure with parameter F = StdErr^
    to output a warning message,
    by default it only writes this message in file F}
  WriteString: WriteProcedureType;
    {BigInt.Write calls this procedure to output a string into the file F}
  ReadString: ReadProcedureType;
    {BigInt.Read calls this procedure to input a string from the file F}
    {By default reads a string using #9, #10, #13, #32 as delimeters}

{----------------------------------------------------------------------------}

implementation

constructor BigInt.Init;
begin
  New(Number);
  with Number^ do
  begin
    Sign := 0;
    Data := nil;
    Len := 0;
    Refs := 1;
    MemLen := 0;
  end;
end;

procedure BigInt.AssignZero;
begin
  _NewNumber(0, nil, 0, 0);
end;

procedure BigInt.AssignOne;
begin
  AssignWord(1);
end;

procedure BigInt.AssignMinusOne;
begin
  AssignWord(1);
  Minus(Self, Self);
end;

procedure _DoError(var F: Text; S: String); far;
begin
  writeln(F, S);
end;

procedure _DoFatalError(var F: Text; S: String); far;
begin
  writeln(F, S);
  Halt;
end;

procedure _DoWarning(var F: Text; S: String); far;
begin
  Writeln(F, S);
end;

destructor BigInt.Done;
begin
  _Free;
end;

procedure BigInt._Free;
begin
  if Number = nil then exit;
  with Number^ do
  begin
    Dec(Refs);
    if Refs <> 0 then exit;
    if Data <> nil then
      FreeMem(Data, MemLen*SizeOf(Word));
  end;
  Dispose(Number);
  Number := nil;
end;

procedure BigInt.Assign(B: BigInt); {A := B}
  {make a copy of a number, just increments the reference count}
begin
  if Number = B.Number then exit;
  _Free;
  Inc(B.Number^.Refs);
  Number := B.Number;
end;

function _DoAdd(p1, p2: PWordArray; m, n: word): word; assembler;
asm
        push ds
        mov cx, n
        mov dx, m
        sub dx, cx

        lds si, p1
        les di, p2

        xor bx, bx              {bx := 0}

        clc                     {clear 䫠 ७}

@loop1: mov ax, es:[di+bx]
        adc [si+bx], ax
        inc bx                  {invalid: add bx, 2}
        inc bx
        loop @loop1

        mov cx, dx              {cx := m-n}
        jcxz @lastfig           {if m=n then done}

        mov dx, 0               {invalid: xor dx, dx}

@loop2: adc [si+bx], dx
        inc bx
        inc bx
        loop @loop2

@lastfig:
        jae @end
        mov dx, 1
        mov [si+bx], dx
        inc bx
        inc bx

@end:   shr bx, 1
        mov ax, bx              {ax := bx := bx/2 the number of "figures"}
        pop ds

end;

function _DoSub(p1, p2: PWordArray; m, n: word): boolean; assembler;
asm
        push ds
        mov cx, n
        mov dx, m
        sub dx, cx

        lds si, p1
        les di, p2

        xor bx, bx              {bx := 0}

        clc                     {clear cf}

@loop1: mov ax, es:[di+bx]
        sbb [si+bx], ax
        inc bx                  {invalid: add bx, 2}
        inc bx
        loop @loop1

        mov cx, dx              {cx := m-n}
        mov ax, 0               {invalid: xor dx, dx}
        jcxz @lastfig           {if m=n then done}

@loop2: sbb [si+bx], ax
        inc bx
        inc bx
        loop @loop2

@lastfig:
        jae @end                {if  뫮  then return ax = 0}
        mov ax, 1               {else return 1}

@end:   pop ds

end;

function _Optimize(a: PWordArray; n: word): word; assembler;
asm
        push ds
        std
        lds si, a
        mov bx, n
        mov cx, bx
        dec bx
        shl bx, 1
        mov ax, ds
        mov es, ax
        mov di, si
        add di, bx
        xor ax, ax
        repe scasw      {find non-zero}
        je @end         {if non found then exit}
        mov ax, cx
        inc ax
@end:   pop ds
end;

procedure Add(var A: BigInt; B, C: BigInt); {A := B + C}
var
  Sum, NewSum, BData, CData: PWordArray;
  SumSign, BSign, CSign: ShortInt;
  SumLen, SumMemLen, BLen, CLen: Word;
begin
  BSign := B.Number^.Sign;
  CSign := C.Number^.Sign;
  BLen := B.Number^.Len;
  CLen := C.Number^.Len;
  BData := B.Number^.Data;
  CData := C.Number^.Data;

  if BSign = 0 then
    A.Assign(C)
  else if CSign = 0 then
    A.Assign(B)
  else
  begin
    if BSign = CSign then
    begin
      SumSign := BSign;
      if BLen > CLen then
      begin
        SumMemLen := BLen + 1;
        GetMem(Sum, SumMemLen * SizeOf(Word));
        Move(BData^, Sum^, BLen * SizeOf(Word));
        SumLen := _DoAdd(Sum, CData, BLen, CLen);
      end
      else
      begin
        SumMemLen := CLen + 1;
        GetMem(Sum, SumMemLen * SizeOf(Word));
        Move(CData^, Sum^, CLen * SizeOf(Word));
        SumLen := _DoAdd(Sum, BData, CLen, BLen);
      end;
    end
    else  {BSign <> CSign}
    begin
      if BLen > CLen then
      begin
        GetMem(Sum, BLen * SizeOf(Word));
        Move(BData^, Sum^, BLen * SizeOf(Word));
        _DoSub(Sum, CData, BLen, CLen);
        SumMemLen := _Optimize(Sum, BLen);
        SumLen := SumMemLen;
        GetMem(NewSum, SumMemLen * SizeOf(Word));
        Move(Sum^, NewSum^, SumMemLen * SizeOf(Word));
        FreeMem(Sum, BLen * SizeOf(Word));
        Sum := NewSum;
        SumSign := BSign;
      end
      else if BLen < CLen then
      begin
        GetMem(Sum, CLen * SizeOf(Word));
        Move(CData^, Sum^, CLen * SizeOf(Word));
        _DoSub(Sum, BData, CLen, BLen);
        SumMemLen := _Optimize(Sum, CLen);
        SumLen := SumMemLen;
        GetMem(NewSum, SumMemLen * SizeOf(Word));
        Move(Sum^, NewSum^, SumMemLen * SizeOf(Word));
        FreeMem(Sum, CLen * SizeOf(Word));
        Sum := NewSum;
        SumSign := CSign;
      end
      else  {BLen = CLen}
      begin
        GetMem(Sum, BLen * SizeOf(Word));
        Move(BData^, Sum^, BLen * SizeOf(Word));
        if _DoSub(Sum, CData, BLen, CLen) then
        begin { }  {B < C}
          Move(CData^, Sum^, CLen * SizeOf(Word));
          _DoSub(Sum, BData, CLen, BLen);
          SumMemLen := _Optimize(Sum, CLen);
          SumLen := SumMemLen;
          GetMem(NewSum, SumMemLen * SizeOf(Word));
          Move(Sum^, NewSum^, SumMemLen * SizeOf(Word));
          FreeMem(Sum, CLen * SizeOf(Word));
          Sum := NewSum;
          SumSign := CSign;
        end
        else
        begin { 뫮 } {B > C}
          SumMemLen := _Optimize(Sum, BLen);
          SumLen := SumMemLen;
          if SumLen = 0 then
          begin
            SumSign := 0;
            NewSum := nil;
          end
          else
          begin
            GetMem(NewSum, SumMemLen * SizeOf(Word));
            Move(Sum^, NewSum^, SumMemLen * SizeOf(Word));
            SumSign := BSign;
          end;
          FreeMem(Sum, BLen * SizeOf(Word));
          Sum := NewSum;
        end;
      end;
    end;
    A._NewNumber(SumSign, Sum, SumLen, SumMemLen);
  end;
end;

procedure Sub(var A: BigInt; B, C: BigInt); {A := B - C}
var
  Sum, NewSum, BData, CData: PWordArray;
  SumSign, BSign, CSign: ShortInt;
  SumLen, SumMemLen, BLen, CLen: Word;
begin
  BSign := B.Number^.Sign;
  CSign := C.Number^.Sign;
  BLen := B.Number^.Len;
  CLen := C.Number^.Len;
  BData := B.Number^.Data;
  CData := C.Number^.Data;

  if CSign = 0 then
    A.Assign(B)
  else if BSign = 0 then
  begin
    if (A.Number = C.Number) and (A.Number^.Refs = 1) then
      A.Number^.Sign := -C.Number^.Sign
    else
    begin
      GetMem(Sum, CLen*SizeOf(Word));
      Move(CData^, Sum^, CLen*SizeOf(Word)); {copy data}
      A._NewNumber(-C.Number^.Sign, Sum, CLen, CLen);
    end;
  end
  else
  begin
    if BSign <> CSign then
    begin
      SumSign := BSign;
      if BLen > CLen then
      begin
        SumMemLen := BLen + 1;
        GetMem(Sum, SumMemLen * SizeOf(Word));
        Move(BData^, Sum^, BLen * SizeOf(Word));
        SumLen := _DoAdd(Sum, CData, BLen, CLen);
      end
      else
      begin
        SumMemLen := CLen + 1;
        GetMem(Sum, SumMemLen * SizeOf(Word));
        Move(CData^, Sum^, CLen * SizeOf(Word));
        SumLen := _DoAdd(Sum, BData, CLen, BLen);
      end;
    end
    else  {BSign = CSign}
    begin
      if BLen > CLen then
      begin
        GetMem(Sum, BLen * SizeOf(Word));
        Move(BData^, Sum^, BLen * SizeOf(Word));
        _DoSub(Sum, CData, BLen, CLen);
        SumMemLen := _Optimize(Sum, BLen);
        SumLen := SumMemLen;
        GetMem(NewSum, SumMemLen * SizeOf(Word));
        Move(Sum^, NewSum^, SumMemLen * SizeOf(Word));
        FreeMem(Sum, BLen * SizeOf(Word));
        Sum := NewSum;
        SumSign := BSign;
      end
      else if BLen < CLen then
      begin
        GetMem(Sum, CLen * SizeOf(Word));
        Move(CData^, Sum^, CLen * SizeOf(Word));
        _DoSub(Sum, BData, CLen, BLen);
        SumMemLen := _Optimize(Sum, CLen);
        SumLen := SumMemLen;
        GetMem(NewSum, SumMemLen * SizeOf(Word));
        Move(Sum^, NewSum^, SumMemLen * SizeOf(Word));
        FreeMem(Sum, CLen * SizeOf(Word));
        Sum := NewSum;
        SumSign := -CSign;
      end
      else  {BLen = CLen}
      begin
        GetMem(Sum, BLen * SizeOf(Word));
        Move(BData^, Sum^, BLen * SizeOf(Word));
        if _DoSub(Sum, CData, BLen, CLen) then
        begin { }  {|B| < |C|}
          Move(CData^, Sum^, CLen * SizeOf(Word));
          _DoSub(Sum, BData, CLen, BLen);
          SumMemLen := _Optimize(Sum, CLen);
          SumLen := SumMemLen;
          GetMem(NewSum, SumMemLen * SizeOf(Word));
          Move(Sum^, NewSum^, SumMemLen * SizeOf(Word));
          FreeMem(Sum, CLen * SizeOf(Word));
          Sum := NewSum;
          SumSign := -CSign;
        end
        else
        begin { 뫮 } {|B| > |C|}
          SumMemLen := _Optimize(Sum, BLen);
          SumLen := SumMemLen;
          if SumLen = 0 then
          begin
            SumSign := 0;
            NewSum := nil;
          end
          else
          begin
            GetMem(NewSum, SumMemLen * SizeOf(Word));
            Move(Sum^, NewSum^, SumMemLen * SizeOf(Word));
            SumSign := BSign;
          end;
          FreeMem(Sum, BLen * SizeOf(Word));
          Sum := NewSum;
        end;
      end;
    end;
    A._NewNumber(SumSign, Sum, SumLen, SumMemLen);
  end;
end;

function _DoMult(u, v, w: PWordArray; m, n: Word): Word; assembler;
var
  vj: Word;
asm
        push ds

        cld             {forward scanning}
        les di, w
        mov cx, m
        xor ax, ax
        rep stosw

        shl m, 1        {m := 2 * m}
        shl n, 1        {n := 2 * n}

        les di, w
        lds si, u

        xor bx, bx      {bx=j := 0}

@loopj:
        push si         {save ds:si -> u}
        push ds
        lds si, v

        mov ax, [si+bx] {save vj}
        mov vj, ax

        pop ds          {retake ds:si -> u}
        pop si
        push bx         {save j}
        xor cx, cx      {p=cx := 0 ७}
        mov bx, cx      {i=bx := 0}

@loopi:
        mov ax, ds:[si+bx]      {ax := u_i}
        mul vj                  {x:y = dx:ax := u_i * v_j}

        add es:[di+bx], cx      {w_i+j := w_i+j + p}
        adc dx, 0               {+ ७}

        add es:[di+bx], ax      {w_i+j := w_i+j + y}
        adc dx, 0               {+ ७}

        mov cx, dx              {७}

        inc bx
        inc bx

        cmp bx, m

        jb @loopi               {if i < m then loop}

        mov es:[di+bx], cx      {w_m+j := p}

        inc di                  {es:di -> w_j+0}
        inc di
        pop bx                  {j}
        inc bx
        inc bx

        cmp bx, n

        jb @loopj               {if j < 2n then loop}

        mov ax, cx              {return major digit}

        pop ds
end;

procedure Mult(var A: BigInt; B, C: BigInt); {A := B * C}
var
  Result: PWordArray;
  BLen, CLen, ResultLen, ResultMemLen: Word;
  BSign, CSign, ResultSign: ShortInt;

begin
  BLen := B.Number^.Len;
  CLen := C.Number^.Len;
  BSign := B.Number^.Sign;
  CSign := C.Number^.Sign;

  if (BSign = 0) or (CSign = 0) then
    A.AssignZero
  else
  begin
    ResultMemLen := BLen + CLen;
    ResultLen := ResultMemLen;
    GetMem(Result, ResultMemLen * SizeOf(Word));
    if _DoMult(B.Number^.Data, C.Number^.Data, Result, BLen, CLen) = 0 then
      Dec(ResultLen);
    A._NewNumber(BSign*CSign, Result, ResultLen, ResultMemLen);
  end;
end;

function  _DoMultByWord(a, p: PWordArray; n, d: word): word; assembler;
asm
        push ds
        lds si, a
        les di, p

        shl n, 1

        xor cx, cx
        mov bx, cx

@loop:  mov ax, ds:[si+bx]
        mul d
        add ax, cx
        adc dx, 0
        mov cx, dx
        mov es:[di+bx], ax

        inc bx
        inc bx

        cmp bx, n

        jb @loop

        mov es:[di+bx], cx
        mov ax, cx

        pop ds
end;

procedure MultByWord(var A: BigInt; B: BigInt; C: Word); {A := B * C}
var
  Result: PWordArray;
  BLen, ResultLen, ResultMemLen: Word;
  BSign, ResultSign: ShortInt;
begin
  BLen := B.Number^.Len;
  BSign := B.Number^.Sign;

  if (BSign = 0) or (C = 0) then
    A.AssignZero
  else
  begin
    ResultMemLen := BLen + 1;
    ResultLen := ResultMemLen;
    GetMem(Result, ResultMemLen * SizeOf(Word));
    if _DoMultByWord(B.Number^.Data, Result, BLen, C) = 0 then
      Dec(ResultLen);
    A._NewNumber(BSign, Result, ResultLen, ResultMemLen);
  end;
end;

function _DoDivideByWord(a, p: PWordArray; n, d: word; var m: Word): word; assembler;
asm
        push ds
        lds si, a
        les di, p
        mov cx, d
        mov bx, n
        dec bx
        shl bx, 1

        xor dx, dx
@loop:  mov ax, ds:[si+bx]
        div cx                  {ax := dx:ax div cx; dx := dx:ax mod cx}
        mov es:[di+bx], ax

        sub bx, 2

        jnc @loop

        lds si, m               {optimize}
        mov bx, n
        mov ds:[si], bx           {m=n}
        dec bx
        shl bx, 1
        mov ax, es:[di+bx]
        cmp ax, 0
        jne @1
        mov bx, ds:[si]
        dec bx
        mov ds:[si], bx           {m=n-1}

@1:     mov ax, dx              {return residue}
        pop ds
end;

procedure DivideByWord(var A: BigInt; B: BigInt; C: Word; var Res: Word);
                                          {A := B div C; R := B mod C}
{For positive B it holds B = A * C + Res}
{For negative B it holds B = A * C - Res}
var
  Q: PWordArray;
  ALen, BLen: Word;
  BSign: ShortInt;
begin
  BLen := B.Number^.Len;
  BSign := B.Number^.Sign;

  if C = 0 then
    FatalError(StdErr^, 'Divide by zero')
  else if BSign = 0 then
  begin
    A.AssignZero;
    Res := 0;
  end
  else
  begin
    GetMem(Q, BLen * SizeOf(Word));
    Res := _DoDivideByWord(B.Number^.Data, Q, BLen, C, ALen);
    if ALen <> 0 then
      A._NewNumber(BSign, Q, ALen, BLen)
    else
    begin
      A.AssignZero;
      FreeMem(Q, BLen * SizeOf(Word));
    end;
  end;
end;

procedure BigInt._FillGetMem(ASign: ShortInt; ALen, AMemLen: Word);
begin
  _Free;
  New(Number);
  with Number^ do
  begin
    Sign := ASign;
    MemLen := AMemLen;
    Len := ALen;
    GetMem(Data, AMemLen * SizeOf(Word));
    Refs := 1;
  end;
end;

procedure BigInt._NewNumber(ASign: ShortInt; AData:PWordArray; ALen, AMemLen: Word);
begin
  _Free;
  New(Number);
  with Number^ do
  begin
    Sign := ASign;
    Data := AData;
    Len := ALen;
    Refs := 1;
    MemLen := AMemLen;
  end;
end;

function _DoDiv(u, v, q: PWordArray; m, n: Word; var l: Word): word; assembler;
var two_n, two_m, v_n_minus_one, v_n_minus_two, d, q_hat, r_hat: Word;
asm
     push ds

     les di, u
     lds si, v

     mov ax, n
     shl ax, 1
     mov two_n, ax
     mov ax, m
     shl ax, 1
     mov two_m, ax

     mov bx, n
     dec bx
     shl bx, 1
     mov ax, ds:[si+bx]
     mov v_n_minus_one, ax

{**********************************************************************}

     {d1: Find d and normilize}

     mov dx, 1
     xor ax, ax                 {b}
     mov cx, v_n_minus_one
     add cx, 1
     jc @j1
     div cx
     mov d, ax
     jmp @j2
@j1: mov d, dx

@j2:
     {multiply v by d}

     xor cx, cx                  {carry}
     mov bx, cx                  {count}

@loop_d1_1:
     mov ax, ds:[si+bx]
     mul d
     add ax, cx
     adc dx, 0
     mov cx, dx
     mov ds:[si+bx], ax
     inc bx
     inc bx
     cmp bx, two_n
     jb @loop_d1_1


     {multiply u by d}

     xor cx, cx                      {carry}
     mov bx, cx                      {count}

@loop_d1_2:
     mov ax, es:[di+bx]
     mul d
     add ax, cx
     adc dx, 0
     mov cx, dx
     mov es:[di+bx], ax
     inc bx
     inc bx
     cmp bx, two_m
     jb @loop_d1_2
     mov es:[di+bx], cx

{**********************************************************************}

     {d2: beginloop on j=bx}

                        {initialize v_n_minus_one, v_n_minus_two}

     mov bx, n
     dec bx
     shl bx, 1
     mov ax, ds:[si+bx]
     mov v_n_minus_one, ax
     mov ax, ds:[si+bx-2]
     mov v_n_minus_two, ax

     mov bx, m
     sub bx, n
     shl bx, 1                  {bx = 2*(m-n)}
     add di, two_m

{**********************************************************************}

     {d3: calculate q_hat}

@d3:

     mov dx, es:[di]            {u[j+n]}
     mov ax, es:[di-2]          {u[j+n-1]}
     cmp dx, v_n_minus_one
     jae @1                     {u[j+n] >= v[n-1]}
     div v_n_minus_one
     mov r_hat, dx
     mov q_hat, ax
     jmp @2
@1:  mov ax, $FFFF              {q_hat}
     mov dx, es:[di-2]          {u[j+n-1}
     jmp @4
@3:  mov ax, q_hat
     dec ax
     mov dx, r_hat
@4:  mov q_hat, ax
     add dx, v_n_minus_one
     jc @d4
     mov r_hat, dx
@2:  mov ax, q_hat
     mul v_n_minus_two
     cmp dx, r_hat
     jb @d4                      {<}
     ja @3                       {>}
     cmp ax, es:[di-4]
     ja @3                       {>}

{************************************************************************}

     {d4: multiply and subtract}

@d4:
     push bx
     sub di, two_n                {u[m]}

     xor cx, cx                   {carry}
     mov bx, cx                   {count}

@loop_d4:

     mov ax, ds:[si+bx]           {v[bx]}
     mul q_hat
     add ax, cx
     adc dx, 0
     sub es:[di+bx], ax           {u[bx]}
     adc dx, 0
     mov cx, dx                   {new carry}
     inc bx
     inc bx
     cmp bx, two_n
     jb @loop_d4
     sub es:[di+bx], cx

{*************************************************************************}

@d5: jnc @d7

{*************************************************************************}

     {d6: q_hat correction}

     dec q_hat
     mov cx, n                    {count}
     xor bx, bx
     clc                          {carry}

@loop_d6:

     mov ax, ds:[si+bx]           {v[bx]}
     adc es:[di+bx], ax           {u[bx]}
     inc bx
     inc bx
     loop @loop_d6
     xor ax, ax
     adc es:[di+bx], ax

{*************************************************************************}

     {d7: endloop on bx}

@d7:
     add di, two_n
     pop bx

     {record q[j]}

     push es
     push di
     les di, q
     mov ax, q_hat
     mov es:[di+bx], ax
     pop di
     pop es

     {endloop on bx}

     dec di
     dec di
     sub bx, 2
     jnc @d3

{*********************************************************************}

     {d8: denormalize}

     les di, u
     mov bx, n
     dec bx
     shl bx, 1          {bx=2*(n-1)}
     xor dx, dx
     mov cx, d

@loop_d8:

     mov ax, es:[di+bx]
     div cx
     mov es:[di+bx], ax
     sub bx, 2
     jnc @loop_d8

{***********************************************************************}

     {optimize u}

     std
     lds si, u
     mov bx, n
     mov cx, bx
     dec bx
     shl bx, 1
     mov ax, ds
     mov es, ax
     mov di, si
     add di, bx
     xor ax, ax
     repe scasw      {find non-zero}
     je @end         {if non found then exit}
     mov ax, cx
     inc ax          {return ax = the number of digits in u}

@end:

     {optimize in q one digit}

     mov bx, m
     sub bx, n
     shl bx, 1                  {bx = 2*(m-n)}
     les di, q
     mov dx, 0
     cmp es:[di+bx], dx
     je @h
     inc bx
     inc bx
@h:  shr bx, 1                  {bx = real lengh of q}
     les di, l
     mov es:[di], bx

     pop ds
end;

procedure Divide(var A: BigInt; B, C: BigInt; var Res: BigInt);
  {A := B div C; R := B mod C}
var
  U, V, Q, R: PWordArray;
  ALen, BLen, CLen, RLen, RWord: Word;
  BSign, CSign: ShortInt;
begin
  BLen := B.Number^.Len;
  CLen := C.Number^.Len;
  BSign := B.Number^.Sign;
  CSign := C.Number^.Sign;

  if CSign = 0 then
    FatalError(StdErr^, 'Divide by zero')
  else if BSign = 0 then
  begin
    A.AssignZero;
    Res.AssignZero;
  end
  else if BLen < CLen then
  begin
    A.AssignZero;
    Res.Assign(B);
  end
  else if BLen >= CLen then
    if CLen = 1 then
    begin
      GetMem(Q, BLen * SizeOf(Word));
      RWord := _DoDivideByWord(B.Number^.Data, Q, BLen, C.Number^.Data^[0], ALen);
      if ALen <> 0 then
        A._NewNumber(BSign * CSign, Q, ALen, BLen)
      else
      begin
        A.AssignZero;
        FreeMem(Q, BLen * SizeOf(Word));
      end;
      if RWord = 0 then
        Res.AssignZero
      else
      begin
        GetMem(U, SizeOf(Word));
        U^[0] := RWord;
        Res._NewNumber(BSign, U, 1, 1);
      end;
    end
    else
    begin
      GetMem(U, (BLen + 1) * SizeOf(Word));
      GetMem(V, CLen * SizeOf(Word));
      GetMem(Q, (BLen - CLen + 1) * SizeOf(Word));
      Move(B.Number^.Data^, U^, BLen * SizeOf(Word));
      Move(C.Number^.Data^, V^, CLen * SizeOf(Word));

      RLen := _DoDiv(U, V, Q, BLen, CLen, ALen);

      FreeMem(V, CLen * SizeOf(Word));
      if RLen = 0 then
      begin
        R := nil;
        Res.AssignZero;
      end
      else
      begin
        GetMem(R, RLen * SizeOf(Word));
        Move(U^, R^, RLen * SizeOf(Word));
        Res._NewNumber(BSign, R, RLen, RLen);
      end;
      FreeMem(U, (BLen + 1) * SizeOf(Word));
      if ALen <> 0 then
        A._NewNumber(BSign*CSign, Q, ALen, BLen - CLen + 1)
      else
      begin
        A.AssignZero;
        FreeMem(Q, (BLen - CLen + 1) * SizeOf(Word));
      end;
    end;
end;

procedure Res(var R: BigInt; B, C: BigInt);
var A: BigInt;
begin
  A.Init;
  Divide(A, B, C, R);
  A.Done;
end;

procedure Dvd(var A: BigInt; B, C: BigInt);
var R: BigInt;
begin
  R.Init;
  Divide(A, B, C, R);
  R.Done;
end;

function _DoBigInt2Bdn(a, b: PWordArray; n: word): Word; assembler;
var two_n_minus_two, f: Word;
asm
        xor ax, ax
        cmp n, ax
        je @exit

        push ds

        lds si, a
        les di, b

        mov f, 10000
        mov bx, n
        dec bx
        shl bx, 1
        mov two_n_minus_two, bx
        xor cx, cx           {count digits in the result}

@next_digit:
        xor dx, dx

@div_loop:
        mov ax, ds:[si+bx]
        div f
        mov ds:[si+bx], ax
        sub bx, 2
        jnc @div_loop

        inc cx               {count digits in the result}
        mov es:[di], dx      {residue: the new digit in the result}
        inc di
        inc di
        mov bx, two_n_minus_two
        xor ax, ax
        cmp ds:[si+bx], ax
        jne @next_digit

        sub bx, 2

        mov two_n_minus_two, bx
        jnc @next_digit

        mov ax, cx            {return the number of digits in the result}

        pop ds

@exit:
end;

function _DoBdn2BigInt(a, b: PWordArray; n: word): Word; assembler;
var f, m, carry: Word;
asm
        push ds

        mov f, 10000

        lds si, a             {first, optimize}

        std

        mov bx, n
        mov cx, bx            {count of digits n}
        dec bx
        shl bx, 1             {2*n-2}
        mov ax, ds
        mov es, ax
        mov di, si
        add di, bx
        xor ax, ax
        repe scasw            {find non-zero}
        mov m, ax             {m=0}
        je @end               {if non found then exit: all digits are zeros}
        inc cx                {the number of non-zero digits}
        inc di
        inc di

        mov si, di
        les di, b

        mov ax, ds:[si]
        mov es:[di], ax
        dec si
        dec si
        mov m, 1
        dec cx
        jcxz @end

@loop:  push cx
        mov cx, m
        xor bx, bx
        mov carry, bx

@mul_loop:
        mov ax, es:[di+bx]
        mul f
        add ax, carry
        adc dx, 0
        mov carry, dx
        mov es:[di+bx], ax
        inc bx
        inc bx
        loop @mul_loop

        cmp carry, 0
        je @h1
        mov ax, carry
        mov es:[di+bx], ax
        inc bx
        inc bx
@h1:    shr bx, 1
        mov cx, bx
        dec cx
        mov bx, 2
        mov ax, ds:[si]
        add es:[di], ax
        jcxz @1

        mov ax, 0                       {invalid: xor ax, ax}
@add_loop:
        adc es:[di+bx], ax
        inc bx
        inc bx
        loop @add_loop
@1:     jnc @h2
        mov ax, 1
        mov es:[di+bx], ax
        inc bx
        inc bx
@h2:    shr bx, 1
        mov m, bx

        dec si
        dec si

        pop cx

        loop @loop

@end:   mov ax, m
        pop ds

end;

function Trim(S : string) : string;
  {Return a string with leading and trailing white space removed}
var
  I : Word;
  SLen : Byte absolute S;
begin
  while (SLen > 0) and (S[SLen] <= ' ') do
    Dec(SLen);

  I := 1;
  while (I <= SLen) and (S[I] <= ' ') do
    Inc(I);
  Dec(I);
  if I > 0 then
    Delete(S, 1, I);

  Trim := S;
end;

function Str2Word(S : string; var I : Word) : Boolean;
  {-Convert a string to a word, returning true if successful}
var
  code : Word;
  SLen : Byte absolute S;
begin
  while S[SLen] = ' ' do
    Dec(SLen);
  Val(S, I, code);
  if code <> 0 then begin
    I := code;
    Str2Word := False;
  end else
    Str2Word := True;
end;

procedure BigInt.AssignString(S: String);
var
  S4: String;
  Len, LenS4: Byte;
  I, J, N, M: Word;
  BDN, Data, NewData: PWordArray;
  Sign: ShortInt;
  FirstS4: Boolean;
begin
  S := Trim(S);
  if S = '' then
  begin
    Warning(StdErr^, 'Empty string. Treating as zero.');
    AssignZero;
  end
  else
  begin
    Sign := 1;
    i := 1;
    Len := Length(S);
    if  S[1] = '+' then i := 2
    else if  S[1] = '-' then
    begin
      Sign := -1;
      i := 2;
    end;
    if i > Len then
    begin
      Warning(StdErr^, 'Error in string. Treating as zero.');
      AssignZero;
    end
    else
    begin
      N := (Len - i + 1) div 4;
      if ((Len - i + 1) mod 4 <> 0) then Inc(N);
      GetMem(BDN, N * SizeOf(Word));
      FirstS4 := True;
      for j := N - 1 downto 0 do
      begin
        if FirstS4 then LenS4 := (Len - i + 1) mod 4 else LenS4 := 4;
        if LenS4 = 0 then LenS4 := 4;
        S4 := Copy(S, i, LenS4);
        if not Str2Word(S4, BDN^[j]) then
          Warning(StdErr^, 'Error has occured in reading a string as a BigInt');
        Inc(i, LenS4);
        FirstS4 := False;
      end;
      GetMem(Data, N * SizeOf(Word));
      M := _DoBdn2BigInt(BDN, Data, N);
      if M = 0 then
      begin
        AssignZero;
        FreeMem(BDN, N * SizeOf(Word));
        FreeMem(Data, N * SizeOf(Word));
      end
      else
      begin
        FreeMem(BDN, N * SizeOf(Word)); {pltcnmz teiipq invalid point operation}
        GetMem(NewData, M * SizeOf(Word));
        Move(Data^, NewData^, M * SizeOf(Word));
        FreeMem(Data, N * SizeOf(Word));
        _NewNumber(Sign, NewData, M, M);
      end;
    end;
  end;
end;

function Long2Str(L : LongInt) : string;
  {Convert an integer number to a string}
var
  S : string;
begin
  Str(L, S);
  Long2Str := S;
end;

function LeftPadCh(S : string; Ch : Char; Len : Byte) : string;
  {Return a string left-padded to length len with ch}
var
  Str : string;
  SLen : Byte absolute S;
begin
  if Length(S) >= Len then
    LeftPadCh := S
  else if SLen < 255 then begin
    Str[0] := Chr(Len);
    Move(S[1], Str[Succ(Word(Len))-SLen], SLen);
    FillChar(Str[1], Len-SLen, Ch);
    LeftPadCh := Str;
  end;
end;

function BigInt.ToString(var S: String): Boolean;
const
  MaxStringLen = 255 div 4;
var
  S4: String[4];
  BDN, NumberData: PWordArray;
  SLen: Byte;
  BDNLen, i: Word;
begin
  if Number^.Sign = 0 then
  begin
    S := '0';
    ToString := True;
    Exit;
  end;
  GetMem(BDN, 2 * Number^.Len * SizeOf(Word));
  GetMem(NumberData, Number^.Len * SizeOf(Word));
  Move(Number^.Data^, NumberData^, Number^.Len * SizeOf(Word));
  BDNLen := _DoBigInt2Bdn(NumberData, BDN, Number^.Len);
  FreeMem(NumberData, Number^.Len * SizeOf(Word));
  if BDNLen >= MaxStringLen then
  begin
    Warning(StdErr^, 'String is too small to contain a BigInt');
    ToString := False;
    S := '';
  end
  else
  begin
    S := Long2Str(BDN^[BDNLen - 1]);
    if BDNLen > 1 then
      for i := BDNLen - 2 downto 0 do
        S := Concat(S, LeftPadCh(Long2Str(BDN^[i]), '0', 4));
    Trim(S);
    if Number^.Sign = -1 then S := Concat('-', S);
    ToString := True;
  end;
  FreeMem(BDN, 2 * Number^.Len * SizeOf(Word));
end;

function Cmp(A, B: BigInt): ShortInt; {Sign(A-B)}
var
  ASign, BSign: ShortInt;
  ALen, BLen: Word;
  C: BigInt;
begin
  ASign := A.Number^.Sign;
  BSign := B.Number^.Sign;
  ALen := A.Number^.Len;
  BLen := B.Number^.Len;

  if ASign < BSign then Cmp := -1
  else if ASign > BSign then Cmp := 1
  else if ASign = 0 then Cmp := 0 {ASign = BSign = 0}
  else if ALen < BLen then Cmp := -ASign {ASign = BSign <> 0}
  else if ALen > BLen then Cmp := ASign
  else {ASign = BSign <> 0, ALen = BLen}
  begin
    C.Init;
    Sub(C, A, B);
    Cmp := C.Number^.Sign;
    C.Done;
  end;
end;

procedure BigInt.AssignWord(W: Word);
var P: PWordArray;
begin
  if W = 0 then
    AssignZero
  else
  begin
    GetMem(P, SizeOf(Word));
    P^[0] := W;
    _NewNumber(1, P, 1, 1);
  end;
end;

function BigInt.IsZero: boolean;
begin
  IsZero := Number^.Sign = 0;
end;

function BigInt.IsPos: boolean;
begin
  IsPos := Number^.Sign = 1;
end;

function BigInt.IsNeg: boolean;
begin
  IsNeg := Number^.Sign = -1;
end;

function BigInt.ToWord(var W: Word): Boolean;
begin
  if Number^.Len > 1 then
  begin
    Warning(StdErr^, 'Word is too small to contain a BigInt');
    ToWord := False;
    W := 0
  end
  else
  begin
    if Number^.Len = 1 then
      W := Number^.Data^[0]
    else
      W := 0;
    ToWord := True;
  end;
end;

function Less(A, B: BigInt): Boolean;
begin
  Less := Cmp(A, B) = -1;
end;

function Equal(A, B: BigInt): Boolean;
begin
  Equal := Cmp(A, B) = 0;
end;

function Greater(A, B: BigInt): Boolean;
begin
  Greater := Cmp(A, B) = 1;
end;

procedure AbsValue(var A: BigInt; B: BigInt);
begin
  if B.IsNeg then
    Minus(A, B)
  else
    A.Assign(B);
end;

procedure Minus(var A: BigInt; B: BigInt); {unary}
var
  Result: PWordArray;
  BLen: Word;
begin
  BLen := B.Number^.Len;
  if B.Number^.Sign = 0 then
    A.AssignZero
  else if (A.Number = B.Number) and (A.Number^.Refs = 1) then
    A.Number^.Sign := -B.Number^.Sign
  else
  begin
    GetMem(Result, BLen * SizeOf(Word));
    Move(B.Number^.Data^, Result^, BLen * SizeOf(Word)); {copy data}
    A._NewNumber(-B.Number^.Sign, Result, BLen, BLen);
  end;
end;

procedure BigInt.BigIntWrite(var F: Text);
const
  MaxStringLen = 255 div 4;
var
  S4: String[4];
  BDN, NumberData: PWordArray;
  SLen: Byte;
  BDNLen, i: Word;
begin
  if Number^.Sign = 0 then
    WriteString(F, '0')
  else
  begin
    if Number^.Sign = -1 then WriteString(F, '-');
    GetMem(BDN, 2 * Number^.Len * SizeOf(Word));
    GetMem(NumberData, Number^.Len * SizeOf(Word));
    Move(Number^.Data^, NumberData^, Number^.Len * SizeOf(Word));
    BDNLen := _DoBigInt2Bdn(NumberData, BDN, Number^.Len);
    FreeMem(NumberData, Number^.Len * SizeOf(Word));
    WriteString(F, Long2Str(BDN^[BDNLen - 1]));
    if BDNLen > 1 then
      for i := BDNLen - 2 downto 0 do
        WriteString(F, LeftPadCh(Long2Str(BDN^[i]), '0', 4));
    FreeMem(BDN, 2 * Number^.Len * SizeOf(Word));
  end;
end;

procedure BigInt.BigIntRead(var F: Text);
var S: String;
begin
  ReadString(F, S);
  AssignString(S);
end;

procedure _DoWriteString(var F: Text; S: String); far;
begin
  Write(F, S);
end;

type
  CharSet = set of Char;

function ReadWord(var F: Text; WordDelims: CharSet) : string;
  {Given a set of word delimiters, return the word from F}
var
  Len : Byte;
  Ch: Char;
begin
  Len := 0;
  ReadWord[0] := #0;

  {skip over delimiters}

  repeat
    Read(F, Ch);
  until not (Ch in WordDelims);

  {we are at the start of a word}

  while (Len < $FF) and not(Ch in WordDelims) do
  begin
    Inc(Len);
    ReadWord[0] := Char(Len);
    ReadWord[Len] := Ch;
    Read(F, Ch);
  end;
end;

procedure _DoReadString(var F: Text; var S: String); far;
begin
  S := ReadWord(F, [#9, #10, #13, #32]);
end;

begin
  Zero.Init;
  One.Init;
  Two.Init;
  MinusOne.Init;

  Zero.AssignZero;
  One.AssignWord(1);
  Two.AssignWord(2);
  Minus(MinusOne, One);

  Error := _DoError;
  FatalError := _DoFatalError;
  Warning := _DoWarning;
  WriteString := _DoWriteString;
  ReadString := _DoReadString;

  StdErr := @Output;
end.
